@@ -1,12 +1,15 @@
 # -*- coding: utf-8 -*-
 
+from SocketServer import TCPServer
 from socket import (AF_INET, SHUT_RDWR, SO_BROADCAST, SO_REUSEADDR, SOCK_DGRAM,
-					SOCK_STREAM, SOL_SOCKET, error, gethostbyname, gethostname,
+					SOCK_STREAM, SOL_SOCKET, error, gethostname, gethostbyname,
 					socket, timeout)
 from time import sleep
 import errno
 import httplib
+import select
 import struct
+import sys
 import threading
 import urllib
 import urlparse
@@ -15,13 +18,29 @@ import demjson
 
 import localization as lang
 from log import safe_print
+from network import get_network_addr
 from util_http import encode_multipart_formdata
 from util_str import safe_unicode
+import webwin
 
 
 _lock = threading.RLock()
 
 
+def Property(func):
+	return property(**func())
+
+
+def _eintr_retry(func, *args):
+	"""restart a system call interrupted by EINTR"""
+	while True:
+		try:
+			return func(*args)
+		except (OSError, select.error) as e:
+			if e.args[0] != errno.EINTR:
+				raise
+
+
 def _shutdown(sock, addr):
 	try:
 		# Will fail if the socket isn't connected, i.e. if there
@@ -29,7 +48,7 @@ def _shutdown(sock, addr):
 		sock.shutdown(SHUT_RDWR)
 	except error, exception:
 		if exception.errno != errno.ENOTCONN:
-			safe_print("MadTPG_Net: SHUT_RDWR for %s:%i failed:" %
+			safe_print("PatternGenerator: SHUT_RDWR for %s:%i failed:" %
 					   addr[:2], exception)
 	sock.close()
 
@@ -52,21 +71,17 @@ class GenHTTPPatternGeneratorClient(object):
 	def __del__(self):
 		self.disconnect_client()
 
-	def _conn_exc(self, exception):
-		msg = lang.getstr("connection.fail", safe_unicode(exception))
-		raise Exception(msg)
-
 	def _request(self, method, url, params=None, headers=None, validate=None):
 		try:
 			self.conn.request(method, url, params, headers or {})
 			resp = self.conn.getresponse()
 		except (error, httplib.HTTPException), exception:
-			self._conn_exc(exception)
+			raise
 		else:
 			if resp.status == httplib.OK:
 				return self._validate(resp, url, validate)
 			else:
-				raise Exception("%s %s" % (resp.status, resp.reason))
+				raise httplib.HTTPException("%s %s" % (resp.status, resp.reason))
 
 	def _shutdown(self):
 		# Override this method in subclass!
@@ -79,7 +94,11 @@ class GenHTTPPatternGeneratorClient(object):
 	def connect(self):
 		self.ip = gethostbyname(self.host)
 		self.conn = httplib.HTTPConnection(self.ip, self.port)
-		self.conn.connect()
+		try:
+			self.conn.connect()
+		except (error, httplib.HTTPException):
+			del self.conn
+			raise
 
 	def disconnect_client(self):
 		self.listening = False
@@ -112,11 +131,10 @@ class GenTCPSockPatternGeneratorServer(object):
 	def wait(self):
 		self.listening = True
 		if self.logfile:
-			host = gethostname()
 			try:
-				host = gethostbyname(host)
-			except socket.error:
-				pass
+				host = get_network_addr()
+			except error:
+				host = gethostname()
 			self.logfile.write(lang.getstr("connection.waiting") +
 							   (" %s:%s\n" % (host, self.port)))
 		while self.listening:
@@ -127,8 +145,7 @@ class GenTCPSockPatternGeneratorServer(object):
 			self.conn.settimeout(1)
 			break
 		if self.listening:
-			if self.logfile:
-				self.logfile.write(lang.getstr("connection.established") + "\n")
+			safe_print(lang.getstr("connection.established"))
 
 	def __del__(self):
 		self.disconnect_client()
@@ -199,6 +216,8 @@ class PrismaPatternGeneratorClient(GenHTTPPatternGeneratorClient):
 		self._event_handlers = {"on_client_added": []}
 		self.broadcast_ip = "255.255.255.255"
 		self.prismas = {}
+		self._size = 10
+		self._enable_processing = True
 
 	def listen(self):
 		self.listening = True
@@ -212,6 +231,8 @@ class PrismaPatternGeneratorClient(GenHTTPPatternGeneratorClient):
 		try:
 			sock.bind(("", port))
 			thread = threading.Thread(target=self._cast_receive_handler,
+									  name="PrismaPatternGeneratorClient.BroadcastHandler[%s:%s]" %
+										   (self.broadcast_ip, port),
 									  args=(sock, self.broadcast_ip, port))
 			self._threads.append(thread)
 			thread.start()
@@ -341,19 +362,19 @@ class PrismaPatternGeneratorClient(GenHTTPPatternGeneratorClient):
 			if "m" in query:
 				method = query["m"][0]
 				if data.get(method) == "Error" and "msg" in data:
-					raise Exception("%s: %s" % (self.host, data["msg"]))
+					raise httplib.HTTPException("%s: %s" % (self.host, data["msg"]))
 			for key, value in validate.iteritems():
 				if key not in data:
-					raise Exception(lang.getstr("response.invalid.missing_key",
+					raise httplib.HTTPException(lang.getstr("response.invalid.missing_key",
 												(self.host, key, raw)))
 				elif value is not None and data[key] != value:
-					raise Exception(lang.getstr("response.invalid.value",
+					raise httplib.HTTPException(lang.getstr("response.invalid.value",
 												(self.host, key, value, raw)))
 			data["raw"] = raw
 			return data
 		elif validate:
 			if raw != validate:
-				raise Exception(lang.getstr("response.invalid",
+				raise httplib.HTTPException(lang.getstr("response.invalid",
 											(self.host, raw)))
 		return raw
 
@@ -365,18 +386,18 @@ class PrismaPatternGeneratorClient(GenHTTPPatternGeneratorClient):
 			win = 1
 		else:
 			win = 2
-		self.invoke("window", "win%i" % win, {"sz": size})
+		self.invoke("Window", "win%i" % win, {"sz": size})
 
 	def get_config(self):
-		return self.invoke("setup", validate={"preset": None, "settings": None,
-											  "vpath": None})
+		return self.invoke("Prisma", "settings", validate={"v": None,
+														   "settings": "Ok"})
 
 	def get_installed_3dluts(self):
-		return self.invoke("cube", "list", validate={"list": "Ok", "tables": None})
+		return self.invoke("Cube", "list", validate={"list": "Ok", "v": None})
 
 	def load_preset(self, presetname):
-		return self.invoke("setup", "loadPreset", {"n": presetname},
-						   validate={"settings": None})
+		return self.invoke("Prisma", "loadPreset", {"n": presetname},
+						   validate={"v": None})
 
 	def load_3dlut_file(self, path, filename):
 		with open(path, "rb") as lut3d:
@@ -389,19 +410,23 @@ class PrismaPatternGeneratorClient(GenHTTPPatternGeneratorClient):
 		self._request("POST", "/fwupload", params, headers)
 
 	def remove_3dlut(self, filename):
-		self.invoke("cube", "remove", {"n": filename})
+		self.invoke("Cube", "remove", {"n": filename})
 
 	def set_3dlut(self, filename):
 		# Select 3D LUT
-		self.invoke("setup", "setCube", {"n": filename, "f": "null"})
+		self.invoke("Prisma", "setCube", {"n": filename, "f": "null"})
 
 	def set_prismavue(self, value):
-		self.invoke("setup", "setPrismaVue", {"a": value, "t": "null"})
+		self.invoke("Prisma", "setPrismaVue", {"a": value, "t": "null"})
 
 	def send(self, rgb=(0, 0, 0), bgrgb=(0, 0, 0), bits=None,
 			 use_video_levels=None, x=0, y=0, w=1, h=1):
 		rgb, bgrgb, bits = self._get_rgb(rgb, bgrgb, bits, use_video_levels)
-		self.invoke("window", "color", {"bg": bgrgb, "fg": rgb})
+		self.invoke("Window", "color", {"bg": bgrgb, "fg": rgb})
+		size = (w + h) / 2.0 * 100
+		if size != self._size:
+			self._size = size
+			self.enable_processing(self._enable_processing, size)
 
 
 class ResolveLSPatternGeneratorServer(GenTCPSockPatternGeneratorServer):
@@ -417,7 +442,7 @@ class ResolveLSPatternGeneratorServer(GenTCPSockPatternGeneratorServer):
 		rgb, bgrgb, bits = self._get_rgb(rgb, bgrgb, bits, use_video_levels)
 		xml = ('<?xml version="1.0" encoding="UTF-8" ?><calibration><shapes>'
 			   '<rectangle><color red="%i" green="%i" blue="%i" />'
-			   '<geometry x="%.2f" y="%.2f" cx="%.2f" cy="%.2f" /></rectangle>'
+			   '<geometry x="%.4f" y="%.4f" cx="%.4f" cy="%.4f" /></rectangle>'
 			   '</shapes></calibration>' % tuple(rgb + [x, y,  w, h]))
 		self.conn.sendall("%s%s" % (struct.pack(">I", len(xml)), xml))
 
@@ -436,10 +461,118 @@ class ResolveCMPatternGeneratorServer(GenTCPSockPatternGeneratorServer):
 		xml = ('<?xml version="1.0" encoding="utf-8"?><calibration>'
 			   '<color red="%i" green="%i" blue="%i" bits="%i"/>'
 			   '<background red="%i" green="%i" blue="%i" bits="%i"/>'
-			   '<geometry x="%.2f" y="%.2f" cx="%.2f" cy="%.2f"/>'
+			   '<geometry x="%.4f" y="%.4f" cx="%.4f" cy="%.4f"/>'
 			   '</calibration>' % tuple(rgb + [bits] + bgrgb + [bits, x, y,
 																  w, h]))
 		self.conn.sendall("%s%s" % (struct.pack(">I", len(xml)), xml))
+
+
+class WebWinHTTPPatternGeneratorServer(TCPServer, object):
+
+	def __init__(self, port, logfile=None):
+		self.port = port
+		Handler = webwin.WebWinHTTPRequestHandler
+		TCPServer.__init__(self, ("", port), Handler)
+		self.timeout = 1
+		self.patterngenerator = self
+		self._listening = threading.Event()
+		self.logfile = logfile
+		self.pattern = "#808080|#808080|0|0|1|1"
+
+	def disconnect_client(self):
+		self.listening = False
+
+	def handle_error(self, request, client_address):
+		safe_print("Exception happened during processing of "
+				   "request from %s:%s:" % client_address,
+				   sys.exc_info()[1])
+
+	@Property
+	def listening():
+		def fget(self):
+			return self._listening.is_set()
+
+		def fset(self, value):
+			if value:
+				self._listening.set()
+			else:
+				self._listening.clear()
+				if hasattr(self, "conn"):
+					self.shutdown_request(self.conn)
+					del self.conn
+				if hasattr(self, "_thread") and self._thread.isAlive():
+					self.shutdown()
+		
+		return locals()
+
+	def send(self, rgb=(0, 0, 0), bgrgb=(0, 0, 0), bits=None,
+			 use_video_levels=None, x=0, y=0, w=1, h=1):
+		pattern = ["#%02X%02X%02X" % tuple(round(v * 255) for v in rgb),
+				   "#%02X%02X%02X" % tuple(round(v * 255) for v in bgrgb),
+				   "%.4f|%.4f|%.4f|%.4f" % (x, y, w, h)]
+		self.pattern = "|".join(pattern)
+
+	def serve_forever(self, poll_interval=0.5):
+		"""Handle one request at a time until shutdown.
+
+		Polls for shutdown every poll_interval seconds. Ignores
+		self.timeout. If you need to do periodic tasks, do them in
+		another thread.
+		"""
+		try:
+			while self._listening.is_set():
+				# XXX: Consider using another file descriptor or
+				# connecting to the socket to wake this up instead of
+				# polling. Polling reduces our responsiveness to a
+				# shutdown request and wastes cpu at all other times.
+				r, w, e = _eintr_retry(select.select, [self], [], [],
+									   poll_interval)
+				if self in r:
+					self._handle_request_noblock()
+		except Exception, exception:
+			safe_print("Exception in WebWinHTTPPatternGeneratorServer.serve_forever:",
+					   exception)
+			self._listening.clear()
+
+	def shutdown(self):
+		"""Stops the serve_forever loop.
+
+		Blocks until the loop has finished. This must be called while
+		serve_forever() is running in another thread.
+		"""
+		self._listening.clear()
+		while self._thread.isAlive():
+			sleep(0.05)
+
+	def wait(self):
+		self.listening = True
+		if self.logfile:
+			try:
+				host = get_network_addr()
+			except error:
+				host = gethostname()
+			self.logfile.write(lang.getstr("webserver.waiting") +
+							   (" %s:%s\n" % (host, self.port)))
+		self.socket.settimeout(1)
+		while self.listening:
+			try:
+				self.conn, addr = self.get_request()
+			except timeout:
+				continue
+			self.conn.settimeout(1)
+			break
+		self.socket.settimeout(None)
+		if self.listening:
+			try:
+				self.process_request(self.conn, addr)
+			except:
+				self.handle_error(self.conn, addr)
+				self.disconnect_client()
+			else:
+				self._thread = threading.Thread(target=self.serve_forever,
+												name="WebWinHTTPPatternGeneratorServerThread")
+				self._thread.start()
+				safe_print(lang.getstr("connection.established"))
 	
 
 if __name__ == "__main__":
